/*
 * thread.h
 *
 * Created: 17.04.2017 12:45:50
 * Author: Detlev Tietjen
 * Licence: LGPL 3.0 
 */

#ifndef THREAD_H_
#define THREAD_H_

#include <stdint.h>
#include <avr/io.h>

///defines the maximum number of threads
#define THREAD_MAX 2
/**
 * Defines the milliseconds per timeslice and must be a positive integer literal.
 * It should be the greatest common divisor of all periods, but must not exceed
 * the value of 261/[CPU-clock in MHz]
 */
#define THREAD_MS_PER_TICK 10
///defines a default status which should returned by the thread function if no state machine is used.
#define THREAD_STATUS_DEFAULT 1

/**
 * forward declaration of the structure holding the data of the thread control block (TCB).
 * The structure itself is capsuled and should only be accessed by setter and getter functions.
 */
struct threadTCBData_t;
///defines the integer type used for the thread state.
typedef int8_t threadStatus_t;

typedef struct threadTCBData_t * pThread_t;

/**
 * Prototype for thread function. The function must return a state value, which is passed
 * back on the next call. This value can be used for a state machine and is not used outside the function,
 * but the return value should never be zero, because this marks the first call of the function.
 * You should use the macro 'THREAD_STATUS_DEFAULT' if you do not use a state machine.
 * @param state actual state of the state machine, if a state machine is used. This value was returned
 * by the last call and is set to zero at the first call of the function.
 * @param pArgument this parameter is defined in threadAdd and is provided to the thread function
 * at every call.
 * @see threadAdd
 */
typedef threadStatus_t (*threadFunc_t)(threadStatus_t state, void * pArgument);

/**
 * Adds a new thread to the list. This function must not be called after starting the scheduling.
 * The priority of the task depends on the order of the adding of the threads. The first
 * function has the highest, the last the lowest priority.
 * @see threadStart
 * 	@param f pointer to the thread function
 * 	@param pArgument this parameter is delivered to the thread function at every call.
 * 	Typically it would be
 * 	a pointer to struct with parameters and/or a pointer to a struct holding the static data of
 * 	the function. (You should avoid using static variables in function used for a thread)
 * 	@param periodTicks number of ticks between to calls. You should use '
 * 	[period in ms]/THREAD_MS_PER_TICK' here.
 * 	@return NULL if the list was already full, not NULL if the task was successfully added.
 */
pThread_t threadAdd(threadFunc_t f, void * pArgument, uint16_t periodTicks);

/**
 * Configures the timer, enables global interrupts and starts executing of the tasks.
 */
void threadStart();

/**
 * Marks the beginning of instructions, which must not interrupted by another thread.
 * This function only affects the timetick of the scheduler, all another interrupts
 * remain active. This block *must* be closed by 'threadEndCriticalBlock'.
 * @see threadEndCriticalBlock
 */
inline void threadBeginCriticalBlock() {
	TIMSK0 = 0;
}

/**
 * Marks the end of instructions, which must not interrupted by another thread.
 * @see threadBeginCriticalBlock
 */
inline void threadEndCriticalBlock() {
	TIMSK0 = (1 << OCIE0A);
}

#endif /* THREAD_H_ */
